(module asplevels-ex-test-exceptions scheme
  (require redex)
  (require "../asplevels-ex-core.rkt")
  (provide asplevels-ex-test-exceptions)
  
  (define (asplevels-ex-test-exceptions λreduction-rules)
    (begin
      (display "Running exceptions tests")
      (newline)      
      
      ;; Testing exceptions
      ; Exception is raised and its level is modified with up
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps)
           (up (raise 1))))
       (term (0 (stack) (asps) (exn 1 1))))
      
      ; Exception is raised and its level is modified with down
      (test-->>
       λreduction-rules
       (term
        (1 (stack) (asps)
           (down (raise 1))))
       (term (1 (stack) (asps) (exn 0 1))))
      
      ; down/up
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps)
           (down (up (raise #t)))))
       (term (0 (stack) (asps) (exn 0 #t))))
      
      ; up/down
      (test-->>
       λreduction-rules
       (term
        (1 (stack) (asps)
           (up (down (raise #t)))))
       (term (1 (stack) (asps) (exn 1 #t))))
      
      ;; Exception propagation tests
      ; if-1
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) (if (raise #f) 1 2)))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; if-2
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) (if #t (raise #f) 0)))
       (term 
        (0 (stack) (asps) (exn 0 #f))))
      
      ; if-3
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) (if #f 0 (raise #f))))
       (term 
        (0 (stack) (asps) (exn 0 #f))))
      
      ; if-4
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) (if #f (raise #f) 0)))
       (term 
        (0 (stack) (asps) 0)))
      
      ; begin-first
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) (begin (raise #f) 1 2 3)))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; begin-middle
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) (begin 1 2 3 (raise #f) 4 5 6)))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; begin-last
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) (begin 1 2 3 (raise #f))))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; raise-raise
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) (raise (raise #f))))
       (term 
        (0 (stack) (asps) (exn 0 #f))))
      
      ; raise-raise...raise
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) (raise (raise (raise (raise (raise (raise (raise #f)))))))))
       (term                              
        (0 (stack) (asps) (exn 0 #f))))
      
      ; function application 1
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) ((λ (x) x) (raise #f))))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; function application 2
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) ((λ (x y z) x) 1 (raise #f) 2)))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; function application 3
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) ((λ (x y z) x) 1 2 (raise #f))))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; function application 1 for level capturing function
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) ((λs (x) x) (raise #f))))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; function application 2 for level capturing function
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) ((λs (x y z) x) 1 (raise #f) 2)))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; function application 3 for level capturing function
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) ((λs (x y z) x) 1 2 (raise #f))))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; exception raised during execution of function
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) ((λ (x) (raise #f)) 1)))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ; exception raised during execution of level capturing function
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps) ((λs (x) (raise #f)) 1)))
       (term
        (0 (stack) (asps) (exn 0 #f))))
      
      ;; Handler tests
      ; No exception is raised
      (test-->>
       λreduction-rules
       (term
        (0 (stack) (asps)
           (try 1 with (λ (x) (+ x 1)))))
       (term (0 (stack) (asps) 1)))
      
      ; Exception is raised and captured at level 0
      (test-->> 
       λreduction-rules
       (term 
        (0 (stack) (asps)
           (try (raise 1) with (λ (x) x))))
       (term (0 (stack) (asps) 1)))
      
      ; Exception is raised at level 1 and not captured at level 0
      (test-->>
       λreduction-rules
       (term 
        (0 (stack) (asps)
           (try (up (raise 1)) with (λ (x) x))))
       (term (0 (stack) (asps) (exn 1 1))))
       
      ; Exception is raised at level 1 and captured at level 1
      (test-->>
       λreduction-rules
       (term 
        (0 (stack) (asps)
           (try (up (raise 1)) with (up (λs (x) x)))))
       (term (0 (stack) (asps) 1)))
      
      ; Exceptions is raised at level 1 with down statement
      ; and handler is installed al level 0 with down statement
      (test-->>
       λreduction-rules
       (term 
        (1 (stack) (asps)
           (try (down (raise 1)) with (down (λs (x) x)))))
       (term (1 (stack) (asps) 1)))      
      
      (test-results))))
  